@@ -1,10 +1,11 @@
 from __future__ import annotations
 
+import glob
 import io
 import json
 import os
 import urllib.parse
-from typing import Dict, Optional, Tuple
+from typing import Dict, List, Optional, Tuple
 
 import pandas as pd
 import requests
@@ -13,6 +14,7 @@ from requests.exceptions import Timeout
 from urllib3.exceptions import InsecureRequestWarning
 
 from morphdb_utils.type import (
+    ListStorageDirectoryResponse,
     LoadDataParams,
     Protocol,
     RefResponse,
@@ -27,10 +29,13 @@ urllib3.disable_warnings(InsecureRequestWarning)
 class MorphApiError(Exception):
     pass
 
-def _get_presigned_url_path(file_path, team_slug):
-    file_name = os.path.basename(file_path)
-    last_dir = os.path.basename(os.path.dirname(file_path))
-    return "morph-storage://resources/fsLog/{}/{}/{}".format(team_slug, last_dir, file_name)
+
+def _get_presigned_url_path(file_path, team_slug, database_id):
+    if file_path.startswith("morph-storage://"):
+        return file_path
+    return "morph-storage://vm/{}/{}/{}".format(
+        team_slug, database_id, file_path if not file_path.startswith("/") else file_path[1:]
+    )
 
 
 def _handle_morph_response(response: requests.Response):
@@ -304,7 +309,11 @@ def ref(reference: str) -> RefResponse:
 
     # 当該リファレンスファイルの情報を取得
     canvas_cell_conf = next(
-        (cell for cell in canvas_json_data["cells"] if cell["filename"] == reference),
+        (
+            cell
+            for cell in canvas_json_data["cells"]
+            if cell["filename"] == os.path.basename(reference)
+        ),
         None,
     )
     if canvas_cell_conf is None:
@@ -329,7 +338,9 @@ def ref(reference: str) -> RefResponse:
         parent_cells=canvas_cell_conf["parentCells"],
         connection_type=canvas_cell_conf["connectionType"],
         connection_slug=canvas_cell_conf["connectionSlug"],
-        storage_path=canvas_cell_conf["storagePath"] if "storagePath" in canvas_cell_conf else None,
+        storage_path=canvas_cell_conf["storagePath"]
+        if "storagePath" in canvas_cell_conf
+        else None,
     )
 
 
@@ -412,7 +423,7 @@ def update_records(*args, **kwargs) -> None:
 
 def load_file_data(*args, **kwargs) -> pd.DataFrame:
     if args and isinstance(args[0], RefResponse):
-        if args[0].cell_type != "readonlySheet" or args[0].storage_path is None:
+        if args[0].cell_type != "file" or args[0].storage_path is None:
             raise MorphApiError(f"Cell {args[0].cell_name} is not a Sheet cell")
         ref_dict = {
             "filename": args[0].storage_path,
@@ -430,7 +441,7 @@ def load_data(*args: LoadDataParams, **kwargs) -> pd.DataFrame:
                 "connection_slug": args[0].connection_slug,
             }
             return _execute_sql_impl(**ref_dict, **kwargs)
-        elif args[0].cell_type == "readonlySheet":
+        elif args[0].cell_type == "file":
             ref_dict = {
                 "filename": args[0].storage_path,
             }
@@ -438,12 +449,15 @@ def load_data(*args: LoadDataParams, **kwargs) -> pd.DataFrame:
         elif args[0].cell_type == "python":
             config_from_env = _read_configuration_from_env()
             ref_file_dir = os.path.join(
-                os.path.dirname(args[0].canvas_json_path), os.path.join("outputs", os.path.splitext(args[0].filename)[0])
+                os.path.dirname(args[0].canvas_json_path),
+                os.path.join("data", "output_" + os.path.splitext(args[0].filename)[0] + "_" + os.path.splitext(args[0].filename)[1].replace(".", "")),
             )
-            file_list = os.listdir("/Users/ts/Queue/Project/Morph/morphdb-utils/tests")
+            file_list = os.listdir(ref_file_dir)
             if len(file_list) < 1:
                 raise MorphApiError(f"Cell {args[0].cell_name} has no output file")
-            storage_path = _get_presigned_url_path(os.path.join(ref_file_dir, file_list[0]), config_from_env["team_slug"])
+            storage_path = _get_presigned_url_path(
+                os.path.join(ref_file_dir, file_list[0]), config_from_env["team_slug"], config_from_env["database_id"]
+            )
             ref_dict = {
                 "filename": storage_path,
             }
@@ -461,6 +475,193 @@ def load_data(*args: LoadDataParams, **kwargs) -> pd.DataFrame:
             raise ValueError("Invalid data cell type provided.")
     else:
         raise ValueError("Invalid data cell type provided.")
+    
+
+def read_dir(
+    dir_path = '/',
+    database_id: str | None = None,
+    base_url: str | None = None,
+    team_slug: str | None = None,
+    api_key: str | None = None,  
+) -> List[str]:
+    current_dir = os.path.dirname(os.path.abspath('__file__'))
+    abs_path = os.path.abspath(os.path.join(current_dir, "data", dir_path))
+    dirs = os.listdir(abs_path)
+    storage_dirs = __list_storage_dirs(abs_path, 1, database_id, base_url, team_slug, api_key)
+    if storage_dirs is not None:
+        for storage_dir in storage_dirs.directories:
+            dirs.append(storage_dir.name)
+        for storage_file in storage_dirs.files:
+            dirs.append(storage_file.name)
+    return list(set(dirs))
+
+
+def get_file(
+    filename: str,
+    database_id: str | None = None,
+    base_url: str | None = None,
+    team_slug: str | None = None,
+    api_key: str | None = None,        
+) -> str:
+    config_from_env = _read_configuration_from_env()
+    if database_id is None:
+        database_id = config_from_env["database_id"]
+    if base_url is None:
+        base_url = config_from_env["base_url"]
+    if team_slug is None:
+        team_slug = config_from_env["team_slug"]
+    if api_key is None:
+        api_key = config_from_env["api_key"]
+
+    current_dir = os.path.dirname(os.path.abspath('__file__'))
+    abs_path = os.path.abspath(os.path.join(current_dir, filename)) if not filename.startswith(current_dir) else filename
+    if os.path.exists(abs_path):
+        return abs_path
+    return _get_presigned_url_path(abs_path, team_slug, database_id)
+
+
+def get_data_path(
+    filename: str,
+    specific_filename: str | None = None,
+    database_id: str | None = None,
+    base_url: str | None = None,
+    team_slug: str | None = None,
+    api_key: str | None = None,
+) -> str:
+    """
+    @param filename: The name of the file or directory to be loaded (cell name).
+    @param specific_filename: The specific name under the directory.
+    """
+    config_from_env = _read_configuration_from_env()
+    if database_id is None:
+        database_id = config_from_env["database_id"]
+    if base_url is None:
+        base_url = config_from_env["base_url"]
+    if team_slug is None:
+        team_slug = config_from_env["team_slug"]
+    if api_key is None:
+        api_key = config_from_env["api_key"]
+
+    current_dir = os.path.dirname(os.path.abspath('__file__'))
+    canvas_json_path, canvas_data = _find_canvas_json(current_dir)
+    if canvas_json_path is None or canvas_data is None:
+        raise MorphApiError("canvas.json not found")
+    cell = next(
+        (
+            cell
+            for cell in canvas_data["cells"]
+            if cell["filename"] == (filename.replace(current_dir, "")[1:] if filename.startswith(current_dir) else filename.replace(current_dir, ""))
+        ),
+        None,
+    )
+    if cell is None:
+        raise MorphApiError(f"Cell {filename} not found in canvas.json.")
+    if cell["cellType"] == "file":
+        dir_path = os.path.join(current_dir, os.path.dirname(cell["filename"]))
+        if specific_filename is not None:
+            specific_filepath = os.path.join(dir_path, specific_filename)
+            if os.path.exists(specific_filepath):
+                return specific_filepath
+            return _get_presigned_url_path(specific_filepath, team_slug, database_id)
+        if os.path.exists(os.path.join(current_dir, filename.replace(current_dir, ""))):
+                return os.path.join(current_dir, filename.replace(current_dir, ""))
+        if cell["storagePath"] is not None:
+            return _get_presigned_url_path(cell["storagePath"], team_slug, database_id)            
+        return None
+
+    _, ext = os.path.splitext(filename)
+    if ext == "":
+        dir_path = os.path.join(current_dir, "data", filename)
+        if specific_filename is not None:
+            specific_filepath = os.path.join(dir_path, specific_filename)
+            if os.path.exists(specific_filepath):
+                return specific_filepath
+            return _get_presigned_url_path(specific_filepath, team_slug, database_id)
+        latest_file = __get_latest_file(dir_path)
+        if latest_file is not None:
+            return latest_file
+        storage_data = __list_storage_dirs(dir_path)
+        if storage_data is not None and len(storage_data.files) > 0:
+            return _get_presigned_url_path(storage_data.files[0].path, team_slug, database_id)
+        return None
+    else:
+        base_file_path = os.path.basename(filename)
+        dir_path = os.path.join(current_dir, "data", "output_" + os.path.splitext(base_file_path)[0] + "_" + os.path.splitext(base_file_path)[1].replace(".", ""))
+        if specific_filename is not None:
+            specific_filepath = os.path.join(dir_path, specific_filename)
+            if os.path.exists(specific_filepath):
+                return specific_filepath
+            return _get_presigned_url_path(specific_filepath, team_slug, database_id)
+        latest_file = __get_latest_file(dir_path)
+        if latest_file is not None:
+            return latest_file
+        storage_data = __list_storage_dirs(dir_path)
+        if storage_data is not None and len(storage_data.files) > 0:
+            return _get_presigned_url_path(storage_data.files[0].path, team_slug, database_id)
+        return None
+
+
+def __get_latest_file(dir_path):
+    files = glob.glob(os.path.join(dir_path, '*'))
+    
+    if not files:
+        return None
+    
+    latest_file = max(files, key=os.path.getmtime)
+    
+    return latest_file
+
+
+def __list_storage_dirs(
+    prefix: str,
+    depth: int = 2,
+    database_id: str | None = None,
+    base_url: str | None = None,
+    team_slug: str | None = None,
+    api_key: str | None = None
+):
+    config_from_env = _read_configuration_from_env()
+    if database_id is None:
+        database_id = config_from_env["database_id"]
+    if base_url is None:
+        base_url = config_from_env["base_url"]
+    if team_slug is None:
+        team_slug = config_from_env["team_slug"]
+    if api_key is None:
+        api_key = config_from_env["api_key"]    
+
+    headers = {
+        "X-Api-Key": api_key,
+    }
+
+    split_url = urllib.parse.urlsplit(base_url)
+    request_url = urllib.parse.urljoin(
+        _canonicalize_base_url(base_url),
+        f"{split_url.path}/resource/morph-storage/directory",
+    )
+
+    query_params = {
+        "databaseId": database_id,
+        "depth": depth,
+        "path": prefix,
+    }    
+    request_url += f"?{urllib.parse.urlencode(query_params)}"
+
+    try:
+        response = requests.get(request_url, headers=headers)
+    except Timeout:
+        raise MorphApiError("Process Timeout while executing load_data")
+    except Exception as e:
+        raise MorphApiError(f"{e}")
+
+    response_body = _handle_morph_response(response)
+
+    try:
+        return ListStorageDirectoryResponse(
+            **response_body
+        )
+    except Exception as e:
+        raise MorphApiError(f"load_data error: {e}")    
 
 
 def generate_report(
@@ -481,6 +682,8 @@ def generate_report(
         team_slug = config_from_env["team_slug"]
     if api_key is None:
         api_key = config_from_env["api_key"]
+    if not "dashboard-api" in base_url:
+        base_url = base_url.replace("api", "dashboard-api")
 
     for ref in refs:
         if ref.cell_type != "python":
@@ -503,11 +706,10 @@ def generate_report(
     request = {
         "databaseId": database_id,
         "canvasName": refs[0].canvas_json_path,
-        "fileNames": [ref.filename for ref in refs],        
+        "fileNames": [ref.filename for ref in refs],
         "prompt": prompt,
         "language": language,
     }
-
     try:
         response = requests.post(url, headers=headers, json=request, verify=True)
     except Timeout:
